<?php

namespace IvozDevTools\EntityGeneratorBundle\Doctrine\ValueObject;

use Doctrine\ORM\Mapping\ClassMetadata;
use IvozDevTools\EntityGeneratorBundle\Doctrine\CodeGeneratorUnitInterface;
use IvozDevTools\EntityGeneratorBundle\Doctrine\Entity\EmbeddedGetter;
use IvozDevTools\EntityGeneratorBundle\Doctrine\Entity\EmbeddedProperty;
use IvozDevTools\EntityGeneratorBundle\Doctrine\Entity\EmbeddedSetter;
use IvozDevTools\EntityGeneratorBundle\Doctrine\EntityTypeTrait;
use IvozDevTools\EntityGeneratorBundle\Doctrine\Getter;
use IvozDevTools\EntityGeneratorBundle\Doctrine\ManipulatorInterface;
use IvozDevTools\EntityGeneratorBundle\Doctrine\Property;
use IvozDevTools\EntityGeneratorBundle\Doctrine\Setter;
use IvozDevTools\EntityGeneratorBundle\Doctrine\StringNode;
use IvozDevTools\EntityGeneratorBundle\Doctrine\UseStatement;
use PhpParser\Lexer;
use PhpParser\Node;
use PhpParser\NodeTraverser;
use PhpParser\NodeVisitor;
use PhpParser\Parser;
use Symfony\Bundle\MakerBundle\Doctrine\RelationManyToMany;
use Symfony\Bundle\MakerBundle\Doctrine\RelationManyToOne;
use Symfony\Bundle\MakerBundle\Doctrine\RelationOneToMany;
use Symfony\Bundle\MakerBundle\Doctrine\RelationOneToOne;
use Symfony\Bundle\MakerBundle\Str;

/**
 * @internal
 */
final class ValueObjectManipulator implements ManipulatorInterface
{
    use EntityTypeTrait;

    const CLASS_USE_STATEMENT_PLACEHOLDER = '/*__class_use_statements*/';
    const CLASS_ATTRIBUTE_PLACEHOLDER = '/*__class_attributes*/';
    const CLASS_METHOD_PLACEHOLDER = '/*__class_methods*/';
    const CLASS_CONSTRUCTOR_ARGS_PLACEHOLDER = '/*__construct_args*/';
    const CLASS_CONSTRUCTOR_PLACEHOLDER = '/*__construct_body*/';
    const CLASS_EQUALS_BODY = '/*__equals_body*/';
    const CLASS_EQUALS_ATTR = '/*__equals_attribute*/';

    private $parser;
    private $lexer;

    private $sourceCode;
    private $ast;

    /** @var Property[]  */
    private $properties = [];
    /** @var CodeGeneratorUnitInterface[]  */
    private $methods = [];
    /** @var UseStatement[]  */
    private $useStatements = [];


    public function __construct(
        string $sourceCode
    ) {
        $this->lexer = new Lexer\Emulative([
            'usedAttributes' => [
                'comments',
                'startLine', 'endLine',
                'startTokenPos', 'endTokenPos',
            ],
        ]);
        $this->parser = new Parser\Php7($this->lexer);
        $this->sourceCode = $sourceCode;
        $this->ast = $this->parser->parse($sourceCode);
    }

    public function updateSourceCode()
    {
        $leftPad = str_repeat(' ', 4);

        $this->updateClass(
            self::CLASS_USE_STATEMENT_PLACEHOLDER,
            $this->useStatements,
            '',
            "\n"
        );

        $this->updateClass(
            self::CLASS_ATTRIBUTE_PLACEHOLDER,
            $this->properties,
            $leftPad
        );

        $this->updateClassConstructor(
            $leftPad
        );

        $this->updateEquals(
            $leftPad
        );

        $this->updateClass(
            self::CLASS_METHOD_PLACEHOLDER,
            $this->methods,
            $leftPad
        );

        // Remove black lines
        $this->sourceCode = preg_replace('/^[\t|\s]+\n+/m', "\n", $this->sourceCode);
        $this->sourceCode = preg_replace('/\n{2,}(\s*\})/m', "\n$1", $this->sourceCode);

        $this->useStatements = [];
        $this->properties = [];
        $this->methods = [];
    }

    public function getSourceCode(): string
    {
        return $this->sourceCode;
    }

    public function addEntityField(string $propertyName, array $columnOptions, $classMetadata, array $comments = [])
    {
        $columnName = $columnOptions['columnName'] ?? $propertyName;
        $typeHint = $this->getEntityTypeHint($columnOptions['type']);

        if ($typeHint == '\\DateTime') {
            $this->addUseStatementIfNecessary(
                'Ivoz\\Core\\Domain\\Model\\Helper\\DateTimeHelper'
            );
        }

        $nullable = $columnOptions['nullable'] ?? false;
        if ($nullable) {
            $comments[] = '@var ?' . $typeHint;
        } else {
            $comments[] = '@var ' . $typeHint;
        }
        $comments = array_merge(
            $comments,
            $this->buildPropertyCommentLines($columnOptions)
        );
        $defaultValue = $columnOptions['options']['default'] ?? null;

        if ($defaultValue) {
            switch ($typeHint) {
                case 'int':
                    $defaultValue = intval($defaultValue);
                    break;
                case 'float':
                    $defaultValue = floatval($defaultValue);
                    break;
            }
        }

        if ('array' === $typeHint) {
            $defaultValue = [];
        }

        $this->addProperty(
            $propertyName,
            $typeHint,
            $columnName,
            '',
            $comments,
            $defaultValue,
            !$nullable,
        );

        $paramDoc = '@param ' . $typeHint . ' $' . $propertyName;
        if ($nullable) {
            $paramDoc .= ' | null';
        }

        $setterComments = [
            'Set ' . $propertyName,
            '',
            $paramDoc,
        ];

        $this->addSetter(
            $propertyName,
            $typeHint,
            $nullable,
            $classMetadata,
            $setterComments,
            $columnOptions,
        );

        $returnHint = '@return ' . $typeHint;
        if ($nullable) {
            $returnHint .= ' | null';
        }
        $getterComments = [
            'Get ' . $propertyName,
            '',
            $returnHint
        ];

        $this->addGetter(
            $propertyName,
            $typeHint,
            // getter methods always have nullable return values
            // because even though these are required in the db, they may not be set yet
            $nullable,
            $getterComments
        );
    }

    public function addEmbeddedEntity(string $propertyName, string $className, ClassMetadata $classMetadata = null)
    {
    }

    public function addManyToOneRelation(RelationManyToOne $manyToOne, ClassMetadata $classMetadata)
    {
    }

    public function addOneToOneRelation(RelationOneToOne $oneToOne, ClassMetadata $classMetadata)
    {
    }

    public function addOneToManyRelation(RelationOneToMany $oneToMany, ClassMetadata $classMetadata)
    {
    }

    public function addManyToManyRelation(RelationManyToMany $manyToMany, ClassMetadata $classMetadata)
    {
    }

    public function addInterface(string $interfaceName, ClassMetadata $classMetadata = null)
    {
    }

    public function addAccessorMethod(string $propertyName, string $methodName, $returnType, bool $isReturnTypeNullable, array $commentLines = [], $typeCast = null)
    {
    }

    public function addGetter(string $propertyName, $returnType, bool $isReturnTypeNullable, array $commentLines = [])
    {
        $this->methods[] = new Getter(
            $propertyName,
            $returnType,
            $isReturnTypeNullable,
            $commentLines
        );
    }


    public function addEmbeddedGetter(string $propertyName, $returnType, bool $isReturnTypeNullable, array $commentLines = [])
    {
        $this->methods[] = new EmbeddedGetter(
            $propertyName,
            $returnType,
            $isReturnTypeNullable,
            $commentLines
        );
    }

    public function addSetter(
        string $propertyName,
        $type,
        bool $isNullable,
        $classMetadata,
        array $commentLines = [],
        array $columnOptions = [],
        string $visibility = 'protected'
    ) {
        $this->methods[] = new Setter(
            $propertyName,
            $type,
            $isNullable,
            $classMetadata,
            $commentLines,
            $columnOptions,
            $visibility
        );
    }

    public function addEmbeddedSetter(
        string $propertyName,
        $type,
        bool $isNullable,
        $classMetadata,
        array $commentLines = [],
        array $columnOptions = [],
        string $visibility = 'protected'
    ) {
        $this->methods[] = new EmbeddedSetter(
            $propertyName,
            $type,
            $isNullable,
            $commentLines,
            $columnOptions,
            $classMetadata,
            $visibility
        );
    }

    public function addProperty(
        string $name,
        string $typeHint,
        string $columnName,
        string $fkFqdn,
        array $comments = [],
        $defaultValue = null,
        bool $required = false,
    ) {
        $this->properties[] = new Property(
            $name,
            $typeHint,
            $columnName,
            $comments,
            $defaultValue,
            $required,
            $fkFqdn,
            'private'
        );
    }

    public function addEmbeddedProperty(
        string $name,
        string $typeHint,
        string $columnName,
        string $fkFqdn,
        array $comments = [],
        $defaultValue = null,
        bool $required = false,
    ) {
        $this->properties[] = new EmbeddedProperty(
            $name,
            $typeHint,
            $columnName,
            $comments,
            $defaultValue,
            $required,
            $fkFqdn
        );
    }

    private function buildPropertyCommentLines(array $options): array
    {
        $comments = [];
        if ($options['fieldName'] !== $options['columnName']) {
            $comments[] = sprintf(
                'column: %s',
                $options['columnName']
            );
        }

        $columnComment = $options['options']['comment'] ?? null;
        if ($columnComment && strpos($columnComment, '[') === 0) {
            $comments[] = sprintf(
                'comment: %s',
                substr($columnComment, 1, -1)
            );
        }

        return $comments;
    }

    /**
     * @return string The alias to use when referencing this class
     */
    public function addUseStatementIfNecessary(string $class): string
    {
        $shortClassName = Str::getShortClassName($class);
        if ($this->isInSameNamespace($class)) {
            return $shortClassName;
        }

        if (!array_key_exists($class, $this->useStatements)) {
            $this->useStatements[$class] = new UseStatement($class);
        }

        return $shortClassName;
    }

    private function getClassNode(): Node\Stmt\Class_
    {
        /** @var Node\Stmt\Class_|null $node */
        $node = $this->findFirstNode(function ($node) {
            return $node instanceof Node\Stmt\Class_;
        });

        if (!$node) {
            throw new \Exception('Could not find class node');
        }

        return $node;
    }

    private function getNamespaceNode(): Node\Stmt\Namespace_
    {
        /** @var Node\Stmt\Namespace_|null $node */
        $node = $this->findFirstNode(function ($node) {
            return $node instanceof Node\Stmt\Namespace_;
        });

        if (!$node) {
            throw new \Exception('Could not find namespace node');
        }

        return $node;
    }

    /**
     * @return Node|null
     */
    private function findFirstNode(callable $filterCallback)
    {
        $traverser = new NodeTraverser();
        $visitor = new NodeVisitor\FirstFindingVisitor($filterCallback);
        $traverser->addVisitor($visitor);
        $traverser->traverse($this->ast);

        return $visitor->getFoundNode();
    }

    private function isInSameNamespace($class)
    {
        $namespace = substr($class, 0, strrpos($class, '\\'));

        return $this->getNamespaceNode()->name->toCodeString() === $namespace;
    }

    private function updateClassConstructor(string $leftPad): void
    {
        $src = [];
        foreach ($this->properties as $property) {
            $value = $property instanceof EmbeddedProperty
                ? $property->getForeignKeyFqdn() . ' $' . $property->getName()
                : $property->getHint() . ' $' . $property->getName();

            $defaultValue = $property->getDefaultValue();
            if ($defaultValue !== null) {
                switch (gettype($defaultValue)) {
                    case 'string':
                        $value .= $defaultValue
                            ? " = '" . $defaultValue . "'"
                            : " = ''";
                        break;
                    case 'integer':
                        $value .= ' = ' . $defaultValue;
                    case 'array':
                        if (empty($defaultValue)) {
                            $value .= ' = []';
                        }
                    default:
                }
            }

            $src[] = $value;
        }
        $srcStr = implode(",\n" . str_repeat($leftPad, 2), $src);

        $this->updateClass(
            self::CLASS_CONSTRUCTOR_ARGS_PLACEHOLDER,
            [new StringNode($srcStr)],
            '',
            ''
        );

        $src = [];
        foreach ($this->properties as $property) {
            $src[] = sprintf(
                '$this->set%s(%s);',
                Str::asCamelCase($property->getName()),
                '$' . $property->getName()
            );
        }
        $srcStr = implode(
            "\n" . str_repeat($leftPad, 2),
            $src
        );

        $this->updateClass(
            self::CLASS_CONSTRUCTOR_PLACEHOLDER,
            [new StringNode($srcStr)],
            '',
            ''
        );
    }

    private function updateEquals(string $leftPad): void
    {
        $attr = lcfirst($this->getClassNode()->name->toString());
        $this->updateClass(
            self::CLASS_EQUALS_ATTR,
            [new StringNode('self $' . $attr)],
            '',
            ''
        );

        $src = [];
        $lastPropertyKey = array_key_last($this->properties);
        foreach ($this->properties as $property) {
            $getter = 'get' . Str::asCamelCase($property->getName());

            if ($property !== $this->properties[$lastPropertyKey]) {
                $src[] = 'if ($this->' . $getter . '() !== $' . $attr . '->' . $getter .'()) {';
                $src[] = $leftPad . 'return false;';
                $src[] = '}';
            } else {
                $src[] = 'return $this->' . $getter . '() === $' . $attr . '->' . $getter .'()';
            }
        }
        $srcStr = implode("\n" . str_repeat($leftPad, 2), $src);

        $this->updateClass(
            self::CLASS_EQUALS_BODY,
            [new StringNode($srcStr)],
            '',
            ''
        );

        $src = [];
        foreach ($this->properties as $property) {
            $src[] = sprintf(
                '$this->set%s(%s);',
                Str::asCamelCase($property->getName()),
                '$' . $property->getName()
            );
        }
        $srcStr = implode(
            "\n" . str_repeat($leftPad, 2),
            $src
        );

        $this->updateClass(
            self::CLASS_CONSTRUCTOR_PLACEHOLDER,
            [new StringNode($srcStr)],
            '',
            ''
        );
    }

    /**
     * @param CodeGeneratorUnitInterface[] $items
     * @param string $leftPad
     */
    private function updateClass(string $placeholder, array $items, string $leftPad, string $join = "\n\n"): void
    {
        $value = '';
        foreach ($items as $item) {
            $value .=
                $item->toString($leftPad)
                . $join
                . $leftPad;
        }

        $this->sourceCode = str_replace(
            $placeholder,
            $value,
            $this->sourceCode
        );
    }

    public function addMethod(\ReflectionMethod $method, $classMetadata)
    {
        // TODO: Implement addMethod() method.
    }

    public function addConstant(string $name, string $value)
    {
        // TODO: Implement addConstant() method.
    }
}
